unit HotImage;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  PPoints = ^TPoints;
  TPoints = array[0..0] of TPoint;

  THotSpots = class;
  THotImage = class;

  TSpot = class
  private
   FPoints:TList;
   FPolyRgn:integer;
   FHintText:string;
   FID:integer;
   FEnabled:boolean;
   function GetRegion:HRgn;
   procedure ReleaseRegion;
   function GetPoint(Index:integer):Tpoint;
   procedure SetPoint(Index:integer;const APoint:TPoint);
  public
   constructor Create;
   destructor Destroy;override;
   procedure Assign(Source:TSpot);

   //returns pointer to array[0..n] of TPoint
   function GetPointArray:PPoints;

   //IS point in spot?
   function PointInSpot(const APoint:TPoint):boolean;

   //number of points in spot
   function Count:integer;

   //Clear all points
   procedure Clear;

   //Point manipulation
   procedure AddPoint(const APoint:TPoint);
   procedure DeletePoint(Index:Integer);
   procedure InsertPoint(Index:integer;const APoint:TPoint);

   property Points[Index:integer]:TPoint read GetPoint Write SetPoint;
   property HintText:string read FHintText write FHintText;
   property SpotID:integer read FID write FID;
   property Enabled:boolean read FEnabled write FEnabled;
  end;


  THotSpots=class(TPersistent)
  private
   FList:TList;
   FOwner:THotImage;
   function GetSpot(Index:integer):TSpot;
   procedure SetSpot(index:integer;ASpot:TSpot);
   procedure WriteSpots(Stream:TStream);
   procedure ReadSpots(Stream:TStream);
  protected
   procedure DefineProperties(Filer: TFiler);override;
   procedure Assign(Source:TPersistent);override;
  public
   constructor Create(AOwner:THotImage);
   destructor Destroy;override;


   //Streaming routines
   procedure SaveToStream(Stream:TStream);
   procedure LoadFromStream(Stream:TStream);

   //Spot manipulations;
   procedure DeleteSpot(Index:integer);
   function AddSpot(ASpot:TSpot):integer;
   function RemoveSpot(Index:integer):TSpot;
   procedure InsertSpot(Index:integer;ASpot:TSpot);
   procedure MoveSpot(CurrentIndex,NewIndex:integer);
   procedure Clear;

   function GetSpotByID(ASpotID:integer):TSpot;

   //Coordinates routines
   function IndexByPoint(const Point:TPoint):integer;
   function SpotByPoint(const Point:TPoint):TSpot;


   function Count:integer;

   property Spots[Index:integer]:TSpot read GetSpot write SetSpot; default;
   property Owner:THotImage read FOwner;

  end;
  TGetHintTextEvent=procedure (Sender:TObject;Spot:TSpot;var Text:String)of object;
  THotImage = class(TImage)
  private
    { Private declarations }
   FHotSpots:THotSpots;
   FSpotIndex:integer;
   FCommonHint:string;
   FOnPaint:TNotifyEvent;
   FOnGetHintText:TGetHintTextEvent;
   procedure DoPaint;
   function GetCanvas:TCanvas;
  protected
    { Protected declarations }
    procedure CMHitTest(var Message:TCMHitTest);message CM_HITTEST;
    procedure Paint;override;
  public
    { Public declarations }
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;

    function CopyPointArray(PointBuffer:PPoints;Spot:TSpot):boolean;
    function CreateRgnHandle(Spot:TSpot):HRgn;
    function CreateRgnByID(ASpotID:integer):HRgn;

    procedure SetHotSpots(AHotSpots:THotSpots);
    property SpotIndex:integer read FSpotIndex;
    property Canvas:TCanvas read GetCanvas;
  published
    { Published declarations }
    property HotSpots:THotSpots read FHotSpots write SetHotSpots;
    property Hint read FCommonHint write FCommonHint;     //hides original TControl.Hint
    {Events}
    property OnPaint:TNotifyEvent read FOnPaint write FOnPaint;
    property OnGetHintText:TGetHintTextEvent read FOnGetHintText write FOnGetHintText;
  end;


implementation


{*** TSpot ***}
constructor TSpot.Create;
begin
 FPoints:=TList.Create;
 FEnabled:=true;
end;

destructor TSpot.Destroy;
begin
 ReleaseRegion;
 FPoints.Free;
end;

procedure TSpot.ReleaseRegion;
begin
 If FPolyRgn<>0 then DeleteObject(FPolyRgn);
 FPolyRgn:=0;
end;

function TSpot.GetRegion:integer;
begin
 if FPolyRgn=0 then
    FPolyRgn:=CreatePolygonRgn(GetPointArray^,Count,WINDING);
 result:=FPolyRgn;
end;

function TSpot.Count:integer;
begin
 result:=FPoints.Count div 2;
end;

procedure TSpot.Clear;
begin
  ReleaseRegion;
  FPoints.Clear
end;

procedure TSpot.InsertPoint(Index:integer;const APoint:TPoint);
begin
  Index:=index*2;
  ReleaseRegion;
  FPoints.Insert(index,pointer(APoint.Y));
  FPoints.Insert(index,pointer(APoint.X));
end;

function TSpot.GetPoint(Index:integer):TPoint;
begin
 Index:=index*2;
 With result,FPoints do
 begin
  X:=longint(Items[Index]);
  Y:=longint(Items[Index+1]);
 end;
end;

procedure TSpot.SetPoint(Index:integer;const APoint:TPoint);
begin
  Index:=Index*2;
  ReleaseRegion;
  FPoints.Items[index]:=pointer(APoint.X);
  FPoints.Items[index+1]:=pointer(APoint.Y);
end;

procedure TSpot.AddPoint(const APoint:TPoint);
begin
  ReleaseRegion;
  FPoints.Add(Pointer(APoint.X));
  FPoints.Add(Pointer(APoint.Y));
end;





procedure TSpot.DeletePoint(Index:Integer);
begin
  Index:=Index*2;
  ReleaseRegion;
  FPoints.Delete(Index);
  FPoints.Delete(Index);
end;

function TSpot.GetPointArray:PPoints;
begin
     result:=PPoints(@FPoints.List[0])
 
end;

function TSpot.PointInSpot(const APoint:TPoint):boolean;
begin
if Count>2 then
with APoint do
  result:=PtInRegion(GetRegion,X,Y)
  else result:=false;
end;

procedure TSpot.Assign(Source:TSpot);
var i:integer;
begin
 if Source<>nil then
 begin
   Clear;
   With Source.FPoints do
    For i:=0 to Count-1 do
     Self.FPoints.Add(Items[i]);
  HintText:=Source.HintText;
  Enabled:=Source.Enabled;
  SpotID:=Source.SpotID;
 end;
end;




{*** THotSpots ***}
constructor THotSpots.Create(AOwner:THotImage);
begin
 inherited Create;
 FOwner:=AOwner;
 FList:=TList.Create;

end;

destructor THotSpots.Destroy;
begin
 Clear;
 FList.Free;
 inherited Destroy;
end;

procedure THotSpots.Clear;
var
 i:integer;
begin
 For i:=0 to FList.Count-1 do
  TSpot(FList.Items[i]).Free;
 FList.Clear;
end;


function THotSpots.GetSpot(Index:integer):TSpot;
begin
 result:=TSpot(FList.Items[Index]);
end;

procedure THotSpots.SetSpot(index:integer;ASpot:TSpot);
begin
 TSpot(FList.Items[Index]).assign(ASpot);
end;

{function PointToString(const Point:TPoint):string ;
begin
 result:=format('%d,%d',[Point.X,Point.Y]);
end;


function StringToPoint(const S:string):TPoint;
var DelimPos:integer;
begin
 DelimPos:=Pos(',',S);
 With result do
 begin
   X:=StrToIntDef(Copy(S,1,DelimPos-1),0);
   Y:=StrToIntDef(Copy(S,DelimPos+1,Length(s)),0);
 end;
end;
 }

procedure THotSpots.WriteSpots(Stream:TStream);
var Spot:TSpot;

procedure WritePoints;
var
    N:integer;
begin
 With Stream do
 begin
  N:=Spot.Count;
  WriteBuffer(N,Sizeof(N));
  WriteBuffer(Spot.GetPointArray^,sizeof(TPoint)*N);
 end;
end;
var
    i,N,tmp:integer;
begin
  With Stream do
  begin
   N:=FList.Count;
   WriteBuffer(N,Sizeof(N));
   for i:=0 to N-1 do
   begin
    Spot:=Spots[i];
    WriteBuffer(Spot.FID,sizeof(integer));
    WriteBuffer(Spot.FEnabled,sizeof(boolean));
    tmp:=length(Spot.FHintText);
    WriteBuffer(tmp,sizeof(integer));
    if tmp<>0 then
    WriteBuffer(PChar(Spot.FHintText)^,tmp);
    WritePoints;
   end;
  end;
end;

procedure THotSpots.ReadSpots(Stream:TStream);
var
 Spot:TSpot;

procedure ReadPoints;
var
   i,n:integer;
    pt:TPoint;
begin
 With Stream do
 begin
  ReadBuffer(N,sizeof(N));
  for i:=0 to N-1 do
  begin
   ReadBuffer(pt,Sizeof(TPoint));
   Spot.AddPoint(pt);
  end;
 end;
end;
var
 i,n,tmp:integer;
begin
Clear;
 With Stream do
 begin
  ReadBuffer(n,sizeof(N));
  For i:=0 to n-1 do
  begin
   Spot:=TSpot.Create;
   ReadBuffer(Spot.FID,sizeof(integer));
   ReadBuffer(Spot.FEnabled,Sizeof(boolean));
   ReadBuffer(tmp,sizeof(integer));
   SetLength(Spot.FHintText,tmp);
   if tmp<>0 then
    ReadBuffer(PChar(Spot.FHintText)^,tmp);
   ReadPoints;
   AddSpot(Spot);
  end;
 end;
end;



procedure THotSpots.DefineProperties(Filer: TFiler);
begin
 inherited DefineProperties(Filer);
 Filer.DefineBinaryProperty('Spots',ReadSpots,WriteSpots,FList.Count>0);
end;


procedure THotSpots.Assign(Source:TPersistent);
var
  i:integer;
  ASpot:TSpot;
begin
  If (Source<>nil) and (Source is THotSpots) then
  begin
    Clear;
    For i:=0 to THotSpots(Source).FList.Count-1 do
    begin
      ASpot:=TSpot.Create;
      ASpot.Assign(THotSpots(Source).Spots[i]);
      AddSpot(ASpot);
    end;
  end
   else inherited Assign(Source);
end;

procedure THotSpots.DeleteSpot(Index:integer);
begin
 TSpot(FList.Items[Index]).Free;
 FList.Delete(Index);
end;


function THotSpots.AddSpot(ASpot:TSpot):integer;
begin
 result:=FList.Add(ASpot);
end;

function THotSpots.RemoveSpot(Index:integer):TSpot;
begin
 result:=TSpot(FList.Items[Index]);
 FList.Delete(Index);
end;

procedure THotSpots.MoveSpot(CurrentIndex,NewIndex:integer);
begin
  FList.Move(CurrentIndex,NewIndex);
end;

procedure THotSpots.InsertSpot(Index:integer;ASpot:TSpot);
begin
 FList.Insert(Index,ASpot);
end;

function THotSpots.IndexByPoint(const Point:TPoint):integer;
var i:integer;
begin
 result:=-1;
 For i:=0 to FList.Count-1 do
 With Spots[i]do
  IF Enabled and PointInSpot(Point) then
  begin
   result:=i;
   break;
  end;
end;

function THotSpots.GetSpotByID(ASpotID:integer):TSpot;
var i:integer;
    Spot:TSpot;
begin
 result:=nil;
 for i:=0 to FList.Count-1 do
 begin
  Spot:=TSpot(FList.Items[i]);
  if Spot.FID =ASpotID then
  begin
   result:=Spot;
   break;
  end;
 end;
end;


function THotSpots.SpotByPoint(const Point:TPoint):TSpot;
var i:integer;
begin
  i:=IndexByPoint(Point);
  if i>=0 then result:=Spots[i] else
   result:=nil;
end;

function THotSpots.Count:integer;
begin
  result:=FList.Count;
end;




{streaming functions}
const HS_Sign:array[0..3]of char='HMAP';

procedure THotSpots.SaveToStream(Stream:TStream);
procedure WriteSign;
begin
 Stream.WriteBuffer(HS_Sign,4);
end;
begin
 WriteSign;
 WriteSpots(Stream);
end;



procedure THotSpots.LoadFromStream(Stream:TStream);
procedure ReadSign;
var i:integer;
begin
 Stream.ReadBuffer(i,4);
 if longint(HS_Sign)<>i then
  raise EStreamError.Create('HotMap: Invalid Stream format');
end;

begin
  ReadSign;
  ReadSpots(Stream);
end;



{*** THotImage ***}
constructor THotImage.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 FHotSpots:=THotSpots.Create(self);
 FSpotIndex:=-1;
end;

destructor THotImage.Destroy;
begin
 FHotSpots.Free;
 inherited Destroy;
end;

procedure THotImage.SetHotSpots(AHotSpots:THotSpots);
begin
 FHotSpots.Assign(AHotSpots);
end;

function THotImage.CreateRgnHandle(Spot:TSpot):HRgn;
var 
 Buffer:PPoints;
 BufferSize:Integer;
begin

 result:=0;
 if Spot=nil then exit;

 BufferSize:=Spot.Count*sizeof(TPoint);

  //should be at least 3 points to form region
 if BufferSize>sizeof(TPoint)*2 then
 begin
   
 GetMem(Buffer,BufferSize);
 try
  if CopyPointArray(Buffer,Spot) then
    result:=CreatePolygonRgn(Buffer^,Spot.Count,WINDING);
 finally
  FreeMem(Buffer,BufferSize);
 end;
 end;
end;

function THotImage.CreateRgnByID(ASpotID:integer):HRgn;
var i:integer;
    Spot:TSpot;
    TmpRgn:HRgn;
begin
 result:=0;
 For  i:=0 to FHotSpots.Count-1 do
 begin
   Spot:=FHotSpots[i];
   if Spot.SpotID=ASpotID then
   begin
    TmpRgn:=CreateRgnHandle(Spot);
    if result=0 then result:=TmpRgn
    else
    begin
     CombineRgn(result,result,TmpRgn,RGN_OR);
     DeleteObject(TmpRgn);
    end;
   end;
 end;
end;


function THotImage.CopyPointArray(PointBuffer:PPoints;Spot:TSpot):boolean;
var
 i:integer;
 N,D,Offset:TPoint;
 SrcPoints:PPoints;
 PointsCount:integer;
begin

 if (Spot=nil) or(Picture.Graphic=nil) then
 begin
  result:=false;
  exit;
 end;

 With DestRect do
 begin
  Offset.x:=Left;
  N.X:=Right-Left;
  Offset.y:=Top;
  N.Y:=Bottom-Top;
 end;
 With Picture.Graphic do
 begin
  D.X:=Width;
  D.Y:=Height;
 end;
  PointsCount:=Spot.Count-1;
  SrcPoints:=Spot.GetPointArray;
  for i:=0 to PointsCount do
   with SrcPoints[i] do
   begin
    PointBuffer[i].X:=MulDiv(X,N.X,D.X)+Offset.X;
    PointBuffer[i].Y:=MulDiv(Y,N.Y,D.Y)+Offset.Y;
   end;
 result:=true;
end;

procedure THotImage.CMHitTest(var Message:TCMHitTest);

function ConvertPoint(X,Y:integer):TPoint;
var D,S,Offset:TPoint;
begin
if (Picture.Graphic=nil) or (Picture.Graphic.Empty) then
begin
 result.X:=-1;
 result.Y:=-1;
 exit;
end;

 With DestRect do
 begin
  D.X:=Right-Left;
  D.Y:=Bottom-Top;
  Offset:=TopLeft;
 end;
 With Picture.Graphic do
 begin
  S.X:=Width;
  S.Y:=Height;
 end;
  Result.X:=MulDiv(X-Offset.X,S.X,D.X);
  Result.Y:=MulDiv(Y-Offset.Y,S.Y,D.Y);

end;


var i:integer;
    TempStr:string;
    Spot:TSpot;
begin

 With Message do
 begin
  i:=FHotSpots.IndexByPoint(ConvertPoint(XPos,YPos));
  if i<>FSpotIndex then
  begin

    Application.CancelHint;
    if i>=0 then
    begin
      Spot:=FHotSpots[i];
      TempStr:=Spot.FHintText;
      if assigned(FOnGetHintText)then
       FOnGetHintText(Self,Spot,TempStr);
      inherited Hint:=TempStr;
    end
    else
     inherited Hint:=FCommonHint;
   FSpotIndex:=i;

  end;
   result:=ord(i>=0);
 end;

end;

//hack method to get native TControlCanvas

type THackClass=class(TGraphicControl)
                 public
                  property Canvas;
               end;

function THotImage.GetCanvas:TCanvas;
begin
  result:=THackClass(Self).Canvas;
end;

procedure THotImage.DoPaint;
begin
 if assigned (FOnPaint) then FOnPaint(self);
end;

procedure THotImage.Paint;
begin
 inherited Paint;
 DoPaint;
end;

end.
