unit HSEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls,HotImage;

type
  TfrmHSEdit = class(TForm)
    GroupBox1: TGroupBox;
    ScrollBox1: TScrollBox;
    btnOk: TButton;
    btnCancel: TButton;
    GroupBox2: TGroupBox;
    lboxSpots: TListBox;
    btnEditSpot: TButton;
    btnNewSpot: TButton;
    btnDeleteSpot: TButton;
    PaintBox1: TPaintBox;
    btnSave: TButton;
    btnLoad: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    btnClear: TButton;
    procedure PaintBox1Paint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lboxSpotsClick(Sender: TObject);
    procedure btnDeleteSpotClick(Sender: TObject);
    procedure btnEditSpotClick(Sender: TObject);
    procedure btnNewSpotClick(Sender: TObject);
    procedure lboxSpotsDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure lboxSpotsDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure lboxSpotsDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure btnSaveClick(Sender: TObject);
    procedure btnLoadClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
  private
    { Private declarations }
    FGraphic:TGraphic;
    FPrivStr:string;
    FPoints:PPoints;
    FBufferSize:Cardinal;
    FSize:TPoint;
    FNominator:Integer;
    FDenominator:Integer;
    FHotSpots:THotSpots;

    FSelectedSpot:TSpot;

    procedure PaintSpot;
    procedure DrawSpot(ASpot:TSpot);
    procedure FillLB;
    procedure UpdateButtons;
  public
    { Public declarations }
    function Edit(HotSpots:THotSpots;const S:String):boolean;

  end;


{ShowDialog function}
function EditHotSpotsDlg(HotSpots:THotSpots; const S:string):boolean;

implementation
 uses  SpotEdit;
{$R *.DFM}

function EditHotSpotsDlg(HotSpots:THotSpots; const S:string):boolean;
begin
 With  TfrmHSEdit.Create(Application)do
 try
   result:=Edit(HotSpots,s);
 finally
   Free;
 end;
end;



const LB_STR_LEN=24;


function GetFirstWords(const S:String):string;
var I:integer;
    Ch:Char;
begin
  result:='';
  i:=1;
  While (i<=Length(S)) and (i<=LB_STR_LEN) do
  begin
   Ch:=S[i];
   If Ch>#13 then
    result:=result+Ch
   else result:=result+' ';
   inc(i);
  end;
  result:=result+'...';
end;

procedure TfrmHSEdit.UpdateButtons;
var Flag:Boolean;
begin
 Flag:=lBoxSpots.ItemIndex>=0;
 btnEditSpot.Enabled:=Flag;
 btnDeleteSpot.Enabled:=flag;
end;


procedure TfrmHSEdit.FillLB;
var i:integer;
    Spot:TSpot;
begin
 With lboxSpots.Items do
 begin
  BeginUpdate;
  try
   Clear;
   For i:=0 to FHotSpots.Count-1 do
   begin
    Spot:=FHotSpots[i];
    AddObject(GetFirstWords(Spot.HintText),Spot);
   end;
  finally
   EndUpdate;
  end;
 end;

end;


procedure TfrmHSEdit.PaintSpot;
begin
 if FPoints<>nil then
 begin
  With PaintBox1.Canvas do
  begin
   Pen.Width:=2;
   Pen.Color:=clBlack;
   Pen.Style:=psSolid;
   Pen.Mode:=pmNotXor;
   Windows.Polyline(Handle,FPoints^,FBufferSize);
  end;
 end;

end;

procedure TfrmHSEdit.DrawSpot(ASpot:TSpot);
var i:integer;
    tmpPoints:PPoints;
begin
 if FSelectedSpot<>ASpot then
 begin
   //Painting old shape
   PaintSpot;

   FreeMem(FPoints,FBufferSize*SizeOf(TPoint));
   If (ASpot<>nil) and (ASpot.Count>2) then
   begin
    FSelectedSpot:=ASpot;
    FBufferSize:=ASpot.Count+1;
    GetMem(FPoints,FBufferSize*SizeOf(TPoint));
    tmpPoints:=ASpot.GetPointArray;
    For i:=0 to FBufferSize-2 do
    begin
     FPoints[i].X:=MulDiv(tmpPoints[i].X,FNominator,FDenominator);
     FPoints[i].Y:=MulDiv(tmpPoints[i].Y,FNominator,FDenominator);
    end;
    FPoints[FBufferSize-1]:=FPoints[0];

    //Painting New Shape
    PaintSpot;
   end

    else
    begin
     FSelectedSpot:=nil;
     FPoints:=nil;
     FBufferSize:=0;
    end;

 end;
end;



function TfrmHSEdit.Edit(HotSpots:THotSpots;const S:String):boolean;
procedure ScaleSize;
begin
  With FGraphic do
  begin
   IF Width>=Height then
   begin
    FNominator:=PaintBox1.Width;
    FDenominator:=FGraphic.Width;
    FSize.X:=FNominator;
    FSize.Y:=MulDiv(FGraphic.Height,FNominator,FDenominator);
   end
   else
   begin
    FNominator:=PaintBox1.Height;
    FDenominator:=FGraphic.Height;

    FSize.X:=MulDiv(FGraphic.Width,FNominator,FDenominator);
    FSize.Y:=FNominator;
   end;
  end;
end;

begin
  FPrivStr:=S;
  Caption:=S;
  FHotSpots.assign(HotSpots);
  if (not Assigned(HotSpots.Owner))or (not Assigned(HotSpots.Owner.Picture.Graphic))
  or (HotSpots.Owner.Picture.Graphic.Empty )then
  begin
   ShowMessage('Picture property undefined');
   result:=false;
   exit;
  end;
  FillLB;
  FGraphic:=HotSpots.Owner.Picture.Graphic;
  ScaleSize;
  btnSave.Enabled:=FHotSpots.Count>0;
  btnClear.Enabled:=FHotSpots.Count>0;
 result:=ShowModal=mrOk;
 if result then
 HotSpots.Assign(FHotSpots);
end;

procedure TfrmHSEdit.PaintBox1Paint(Sender: TObject);
begin
  PaintBox1.Canvas.StretchDraw(Rect(0,0,FSize.X,FSize.Y),FGraphic);
  PaintSpot;
end;

procedure TfrmHSEdit.FormCreate(Sender: TObject);
begin
 FHotSpots:=THotSpots.Create(nil);
 FSelectedSpot:=nil;
end;

procedure TfrmHSEdit.FormDestroy(Sender: TObject);
begin
 FHotSpots.Free;
 FreeMem(FPoints,FBufferSize*SizeOf(TPoint));
end;

procedure TfrmHSEdit.PaintBox1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  lboxSpots.ItemIndex:=FHotSpots.IndexByPoint(Point(Muldiv(X,FDenominator,FNominator),
                                                    Muldiv(Y,FDenominator,FNominator)));
  lboxSpotsClick(self);
end;

procedure TfrmHSEdit.lboxSpotsClick(Sender: TObject);
var Spot:TSpot;
begin
UpdateButtons;
With lboxSpots do
 If ItemIndex<0 then
 Spot:=nil
 else Spot:=TSpot(Items.Objects[ItemIndex]);
DrawSpot(Spot);

end;

procedure TfrmHSEdit.btnDeleteSpotClick(Sender: TObject);
var i:integer;
begin


With lboxSpots do
 begin
 i:=ItemIndex;
 Items.Delete(i);
 end;
  FHotSpots.DeleteSpot(i);
 lboxSpotsClick(self);
  btnSave.Enabled:=FHotSpots.Count>0;
  btnClear.Enabled:=FHotSpots.Count>0;
end;

procedure TfrmHSEdit.btnEditSpotClick(Sender: TObject);
var Spot:TSpot;
begin
 Spot:=FHotSpots[lboxSpots.ItemIndex];
 With TfrmSpotEdit.Create(Application) do
 try
  Caption:=Format('%s[%d]',[FPrivStr,lboxSpots.ItemIndex]);
  if Edit(FGraphic,Spot)then
  begin
   UpdateWindow(self.ScrollBox1.Handle);
   DrawSpot(nil);
   With  lboxSpots do
    Items[ItemIndex]:=GetFirstWords(Spot.HintText);
   lboxSpotsClick(self);

  end;
 finally
  Free;
 end;

end;

procedure TfrmHSEdit.btnNewSpotClick(Sender: TObject);
var Spot:TSpot;

function GetNewSpotID:integer;
var i:integer;
begin
 result:=0;
 For i:=0 to FHotSpots.Count-1 do
  With FHotSpots[i]do
   if result<SpotID then result:=SpotID;
 inc(result);
end;

begin
 Spot:=TSpot.Create;
 Spot.SpotID:=GetNewSpotID;
 With TfrmSpotEdit.Create(Application) do
 try
   Caption:=Format('%s - New Spot',[FPrivStr]);
  if Edit(FGraphic,Spot)then
  begin
   UpdateWindow(self.ScrollBox1.Handle);
   FHotSpots.AddSpot(Spot);
   //DrawSpot(Spot);
   With  lboxSpots do
   begin
    Items.AddObject(GetFirstWords(Spot.HintText),Spot);
    ItemIndex:=Items.Count-1;
   end;
   lboxSpotsClick(self);
    btnSave.Enabled:=true;
    btnClear.Enabled:=true;
  end
  else
  Spot.Free;
 finally
  Free;
 end;

end;

procedure TfrmHSEdit.lboxSpotsDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);

var Spot:TSpot;
begin

 With lboxSpots do
 begin
  Spot:=TSpot(Items.Objects[Index]);
  if odSelected	in State
  then
  begin
   Canvas.Brush.Color:=clHighlight;
   Canvas.Font.Color:=clHighlightText;
  end
  else
  begin

   if not Spot.Enabled then
    Canvas.Font.Color:=clGrayText;
  end;

   Canvas.FillRect(Rect);
   With Rect do
   begin
    Canvas.TextOut(left+1,Top,IntToStr(Spot.SpotID));
    Canvas.TextOut(60,Top,Items[Index]);
   end;
 end;
end;

procedure TfrmHSEdit.lboxSpotsDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
Accept:=(Sender=Source)and (lboxSpots.ItemIndex>=0);
end;

procedure TfrmHSEdit.lboxSpotsDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var S,D:integer;

begin
 if Sender=Source then
 begin
  With lboxSpots do
  begin
   S:=ItemIndex;
   D:=ItemAtPos(point(X,Y),true);
   if (D>=0) and (S<>D) then
   begin
    Items.Move(S,D);
    FHotSpots.MoveSpot(S,D);
    ItemIndex:=D;
   end;
  end;
 end;
end;

procedure TfrmHSEdit.btnSaveClick(Sender: TObject);
var fs:TFileStream;
begin
 if SaveDialog1.Execute then
 begin
  fs:=TFileStream.Create(SaveDialog1.FileName,fmCreate);
  try
   FHotSpots.SaveToStream(fs);
  finally
   fs.Free;
  end;
 end;
end;

procedure TfrmHSEdit.btnLoadClick(Sender: TObject);
var fs:TFileStream;
begin
  if OpenDialog1.Execute then
 begin
  UpdateWindow(ScrollBox1.Handle);
  DrawSpot(nil);
  fs:=TFileStream.Create(OpenDialog1.FileName,fmOpenRead {or fmShareDenyRead});
  try
   FHotSpots.LoadFromStream(fs);
  finally
   fs.Free;
  end;

  FillLB;
  if FHotSpots.Count>0 then
  begin
   btnSave.Enabled:=true;
   btnClear.Enabled:=true;
   lboxSpots.ItemIndex:=0;
   lboxSpotsClick(self);
  end;
 end;
end;

procedure TfrmHSEdit.btnClearClick(Sender: TObject);
begin
if MessageDlg('Do you really want to remove all Spots?',mtConfirmation,[mbYes,mbNo],0)<>mrYes
 then exit;
 DrawSpot(nil);
 FHotSpots.Clear;
 FillLB;
 btnSave.Enabled:=false;
 btnClear.Enabled:=false;
 lboxSpotsClick(self);
end;

end.
