unit SpotEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,ExtCtrls,HotImage, Buttons;

type

  TfrmSpotEdit = class(TForm)
    GroupBox1: TGroupBox;
    ScrollBox1: TScrollBox;
    GroupBox2: TGroupBox;
    memHint: TMemo;
    Label1: TLabel;
    lboxPoints: TListBox;
    Label2: TLabel;
    btnOk: TButton;
    btnCancel: TButton;
    sbEdit: TSpeedButton;
    sbDeletePoint: TSpeedButton;
    Label3: TLabel;
    chkboxEnabled: TCheckBox;
    edSpotID: TEdit;
    rgZoom: TRadioGroup;
    Label4: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure sbEditClick(Sender: TObject);
    procedure ScrollBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lboxPointsClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure sbDeletePointClick(Sender: TObject);
    procedure ScrollBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ScrollBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure edSpotIDKeyPress(Sender: TObject; var Key: Char);
    procedure rgZoomClick(Sender: TObject);
  private
    { Private declarations }
    FGraphic:TGraphic;

    FPaintRect:TRect;
    FNominator,
    FDenominator:integer;

    FPaintBox:TPaintBox;

    //points that  stores independed coordinates;
    FPoints:PPoints;
    FZmdPoints:PPoints;

    FPointsCount:integer;
    FSelectedPoint:Integer;


    FDrawing:Boolean;
    FOutOfControl:Boolean;
    FHiding:boolean;
    FOldPoint:TPoint;
    FOldSBProc:pointer;
    FNewSBProc:pointer ;
    procedure DoPaint(Sender:TObject);
    procedure FillLB;
    procedure SetSelectedPoint(Index:integer);
    procedure ResortPoints;
    procedure GetPoints(ASpot:TSpot);
    procedure PaintShape;
    procedure DeletePoint(index:integer);
    procedure AddPoint(const APoint:TPoint);
    procedure DrawEditLines;
    procedure HandleSBMessages(var Message:TMessage);

    procedure SetScale(Nominator,Denominator:integer);

    //Scale routines;
    function ItoD(const P:TPoint):TPoint;
    function DtoI(const P:TPoint):TPoint;
  public
    { Public declarations }
    function Edit(Graphic:TGraphic;Spot:TSpot):boolean;
  end;

var
  frmSpotEdit: TfrmSpotEdit;

implementation

{$R *.DFM}
type
  {Internal PaintBox without Mouse event handles}
  TPaintBoxInt=class(TPaintBox)
  protected
    procedure CMHitTest(var Message: TCMHitTest); message CM_HITTEST;
  end;

procedure TPaintBoxInt.CMHitTest(var Message: TCMHitTest);
begin
 Message.Result:=0;
end;

procedure DrawInvRect(const P:TPoint;Canvas:TCanvas);
begin
 With Canvas do
 begin
  Brush.Style:=bsSolid;
  Brush.Color:=clBlack;
 // Pen.Color:=clBlack;
  Pen.Mode:=pmNotXor;
  //Pen.Style:=psSolid;
  With p do
  Rectangle(X-4,Y-4,X+4,Y+4);
 end;
end;

procedure TfrmSpotEdit.HandleSBMessages(var Message:TMessage);
begin
With Message do
begin
 if Msg=WM_CAPTURECHANGED then
 begin
   if FDrawing and not (FOutOfControl)then
   begin
    DrawEditLines;
    FHiding:=true;
   end;
 end
 else
  result:=CallWindowProc(FOldSBProc,ScrollBox1.Handle,Msg,WParam,LParam);
end;
end;

procedure TfrmSpotEdit.ResortPoints;
var buff:PPoints;
    nSize1,nSize2:Cardinal;

begin
 if (FSelectedPoint=-1) and (FPointsCount>0) then
   FSelectedPoint:=FPointsCount-1;
 if (FSelectedPoint<>FPointsCount-1) then
 begin
   nSize1:=sizeof(TPoint)*(FSelectedPoint+1);
   nSize2:=sizeof(TPoint)*FPointsCount-nSize1;

   Buff:=AllocMem(nSize1);
   Move(FPoints^,Buff^,nSize1);
   Move(FPoints[FSelectedPoint+1],FPoints^,nSize2);
   Move(buff^,FPoints[FPointsCount-FSelectedPoint-1],nSize1);
   FreeMem(buff);

   Buff:=AllocMem(nSize1);
   Move(FZmdPoints^,Buff^,nSize1);
   Move(FZmdPoints[FSelectedPoint+1],FZmdPoints^,nSize2);
   Move(buff^,FZmdPoints[FPointsCount-FSelectedPoint-1],nSize1);
   FreeMem(buff);

 end;

end;

procedure TfrmSpotEdit.DeletePoint(index:integer);
var
 nSize:integer;
begin
  PaintShape;
  lboxPoints.Items.Delete(index);
  nSize:=sizeof(TPoint)*(FPointsCount-1);
  if nSize>0 then
  begin
   If index<FPointsCount-1 then
   Move(FPoints[index+1],FPoints[index],nSize-sizeof(TPoint)*index);
   ReallocMem(FPoints,nSize);
   If index<FPointsCount-1 then
   Move(FZmdPoints[index+1],FZmdPoints[index],nSize-sizeof(TPoint)*index);
   ReallocMem(FZmdPoints,nSize);
  end;
   Dec(FPointsCount);
   PaintShape;
end;

procedure TfrmSpotEdit.AddPoint(const APoint:TPoint);
var
 nSize:Integer;
begin
  PaintShape;

  nSize:=sizeof(TPoint)*(FPointsCount+1);
  if nSize>sizeof(TPoint) then
  begin
   ReallocMem(FPoints,nSize);
   ReallocMem(FZmdPoints,nSize);
  end;

  FPoints[FPointsCount]:=APoint;
  FZmdPoints[FPointsCount]:=ItoD(APoint);
  inc(FPointsCount);

  PaintShape;


  FOldPoint:=APoint;
end;

procedure TfrmSpotEdit.SetSelectedPoint(Index:integer);
begin
  if FSelectedPoint>=0 then DrawInvRect(FZmdPoints[FSelectedPoint],FPaintBox.Canvas);
  FSelectedPoint:=Index;
  if Index>=0 then DrawInvRect(FZmdPoints[FSelectedPoint],FPaintBox.Canvas);
end;



procedure TfrmSpotEdit.DrawEditLines;
var TempPoint:TPoint;
begin
 if FPointsCount=0 then exit;

 TempPoint:=ItoD(DtoI(FPaintBox.ScreenToClient(ScrollBox1.ClientToScreen(FOldPoint))));

 with FPaintBox.Canvas do
 begin

 // Brush.Color:=clBlack;
  Brush.Style:=bsClear;
 // Pen.Width:=1;
  Pen.Color:=clBlack;
  Pen.Mode:=pmNotXor;
  Pen.Style:=psDot;

  with FZmdPoints[FPointsCount-1] do
  MoveTo(X,Y);
  With TempPoint do
   LineTo(X,Y);
  if FPointsCount>1 then
   with FZmdPoints[0]do
   LineTo(X,Y);
 end;
end;

procedure TfrmSpotEdit.PaintShape;

begin
With FPaintBox.Canvas do
begin

   Pen.Color:=clBlack;
   Pen.Mode:=pmNotXor;
   Pen.Style:=psSolid;
   Windows.Polyline(Handle,FZmdPoints^,FPointsCount);
   if (not FDrawing) and (FPointsCount>2) then
   begin
    begin

    With FZmdPoints[FPointsCount-1]do
     MoveTo(X,Y);
     With FZmdPoints[0]do
      LineTo(X,Y);
    end;
   end;
end;
end;

procedure TfrmSpotEdit.DoPaint(Sender:TObject);
begin
 with FPaintBox.Canvas do
   StretchDraw(FPaintRect,FGraphic);
  PaintShape;

 if FSelectedPoint>=0 then
  DrawInvRect(FZmdPoints[FSelectedPoint],FPaintBox.Canvas);

end;

procedure TfrmSpotEdit.FillLB;
var i:integer;
function PointToString(const P:TPoint):string;
begin
 result:=format('(%d,%d)',[P.X,P.Y]);
end;

begin
 With lboxPoints.Items do
 begin
  BeginUpdate;
  try
   Clear;
   For i:=0 to FPointsCount-1 do
    Add(PointToString(FPoints[i]));
  finally
   EndUpdate;
  end;
 end;
 with lboxPoints do
  ItemIndex:=Items.Count-1 ;
 lboxPointsClick(self);
end;


procedure TfrmSpotEdit.GetPoints(ASpot:TSpot);
var
 nSize:integer;
begin
 nSize:=ASpot.Count*sizeof(TPoint);
 if nSize>Sizeof(TPoint)  then
 begin
  ReAllocMem(FPoints,nSize);
  ReallocMem(FZmdPoints,nSize)
 end;
 if nSize>0 then
  Move(ASpot.GetPointArray^,FPoints^,ASpot.Count*sizeof(TPoint));
 //Still need to to Recalc FZmdPoints
  FPointsCount:=ASpot.Count;
end;


// Scale conversion for coordinates
function TfrmSpotEdit.ItoD(const P:TPoint):TPoint;
begin
 with p do
 begin
  result.X:=MulDiv(X,FNominator,FDenominator);
  result.Y:=MulDiv(Y,FNominator,FDenominator);
 end;
end;

function TfrmSpotEdit.DtoI(const P:TPoint):TPoint;
begin
 with p do
 begin
  result.X:=MulDiv(X,FDenominator,FNominator);
  result.Y:=MulDiv(Y,FDenominator,FNominator);
 end;
end;


procedure TfrmSpotEdit.SetScale(Nominator,Denominator:integer);

 function Max(X,Y:integer):integer;
 begin
  result:=X;
  if X<Y then result:=Y
 end;

var
 i:integer;
begin
 FNominator:=Nominator;
 FDenominator:=Denominator;
// Cancelling drawing if active
 ReleaseCapture;

 FPaintBox.Parent:=nil;
 //Setting painting rectangle for graphic
 FPaintRect:=Rect(0,0,MulDiv(FGraphic.Width,Nominator,Denominator),
                      MulDiv(FGraphic.Height,Nominator,Denominator));
 With FPaintRect,ScrollBox1 do

  FPaintBox.SetBounds(0,0,Max(Right,ClientWidth),Max(Bottom,ClientHeight));


 //Rescaling points for current Zoom-in ratio
 For i:=0 to FPointsCount-1 do
 begin
  With FPoints[i] do
  begin
   FZmdPoints[i].X:=MulDiv(X,Nominator,Denominator);
   FZmdPoints[i].Y:=MulDiv(Y,Nominator,Denominator);
  end;

 end;

 FPaintBox.Parent:=ScrollBox1;


 UpdateWindow(ScrollBox1.Handle);
end;

function TfrmSpotEdit.Edit(Graphic:TGraphic;Spot:TSpot):boolean;
var i:integer;
begin

 FGraphic:=Graphic;
 GetPoints(Spot);
 SetScale(1,1);

 chkboxEnabled.Checked:=Spot.Enabled;
 edSpotid.Text:=IntToStr(Spot.SpotID);
 memHint.Text:=Spot.HintText;

 FillLB;
 result:= ShowModal=mrOk;

 if result then
  with Spot do
  begin
   HintText:=MemHint.Text;
   Enabled:=chkboxEnabled.Checked;
   SpotID:=StrToIntDef(edSpotID.Text,0);
   Clear;
   for i:=0 to FPointsCount-1 do
   AddPoint(FPoints[i]);
  end;
end;

procedure TfrmSpotEdit.sbEditClick(Sender: TObject);
var SelPoint:integer;
begin
 if sbEdit.Down then
 begin
  ScrollBox1.Cursor:=crCross;
  lboxPoints.Enabled:=false;
  //BeginDrawing
  FOutOfControl:=true;
  PaintShape;
  SelPoint:=FSelectedPoint;
  SetSelectedPoint(-1);
  FSelectedPoint:=SelPoint;
  ResortPoints;
  FSelectedPoint:=-1;
  FDrawing:=true;
  FHiding:=true;
  PaintShape;

 end
   else
   begin
    PaintShape;
    FDrawing:=false;
    PaintShape;
    ScrollBox1.Cursor:=crDefault;
    with lboxPoints do
     Enabled:=true;
    FillLB;
   end;
end;

procedure TfrmSpotEdit.sbDeletePointClick(Sender: TObject);
begin
 SetSelectedPoint(-1);
 DeletePoint(lboxPoints.ItemIndex);
 lboxPointsClick(self);
end;



procedure TfrmSpotEdit.ScrollBox1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 If Button=mbRight then
 begin
  ScrollBox1.Cursor:=crDefault;
  ReleaseCapture;
  SbEdit.Down:=false;
  sbEditClick(self);
 end

end;

procedure TfrmSpotEdit.lboxPointsClick(Sender: TObject);
var P:TPoint;
    R:TRect;
begin
 SetSelectedPoint(lboxPoints.ItemIndex);
  sbDeletePoint.Enabled:=(FSelectedPoint>=0);
//Show Active Selection
  if Sender=lboxPoints then
  begin
   p:=ScrollBox1.ScreenToClient(FPaintBox.ClientToScreen(FZmdPoints[FSelectedPoint]));
   R:=ScrollBox1.ClientRect;
   if not ptInRect(R,p) then
   begin
    if P.X>R.Right then P.X:=P.X-R.Right
    else if P.X>0 then P.X:=0;
    if P.Y>R.Bottom then P.Y:=P.Y-R.Bottom
    else if P.Y>0 then P.Y:=0;

    if P.X>0 then P.X:=P.X+R.Right div 2
    else
    if P.X<0 then P.X:=P.X-R.Right div 2 ;

    if P.Y>0 then P.Y:=P.Y+R.Bottom div 2
    else
     if P.Y<0 then P.Y:=P.Y-R.Bottom div 2 ;

     With ScrollBox1 do
     begin
      HorzScrollBar.Position:=HorzScrollBar.Position+P.x;
      VertScrollBar.Position:=VertScrollBar.Position+P.y;
     end;

   end;

  end;
end;

procedure TfrmSpotEdit.FormCreate(Sender: TObject);
begin
 FSelectedPoint:=-1;
 FPaintBox:=TPaintBoxInt.Create(self);
 FPaintBox.OnPaint:=DoPaint;
 FPaintBox.Parent:=ScrollBox1;

 //Subclassing ScrollBox's Window
 FOldSBProc:=pointer(GetWindowLong(ScrollBox1.Handle,GWL_WNDPROC));
 FNewSBProc:=MakeObjectInstance(HandleSBMessages);
 SetWindowLong(ScrollBox1.Handle,GWL_WNDPROC,longint(FNewSBProc));

 //Allocating memory for point arrays
 FPoints:=AllocMem(sizeof(TPoint));
 FZmdPoints:=AllocMem(sizeof(TPoint));
end;

procedure TfrmSpotEdit.FormDestroy(Sender: TObject);
begin
 SetWindowLong(ScrollBox1.Handle,GWL_WNDPROC,longint(FOldSBProc));
 FreeObjectInstance(FNewSBProc);
 FreeMem(FPoints);
 FreeMem(FZmdPoints);

end;


procedure TfrmSpotEdit.ScrollBox1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin

 if FDrawing then
 begin
  if not ptInRect(ScrollBox1.ClientRect,Point(X,Y))then
  begin

    ReleaseCapture;
    FOutOfControl:=true;
  end
   else
    begin
     if(GetCapture<>ScrollBox1.Handle) then

      SetCapture(ScrollBox1.Handle);
      FOutOfControl:=false;


      if not FHiding then
        DrawEditLines;


      FOldPoint:=Point(X,Y);
      DrawEditLines;
      FHiding:=false;

    end;
 end;
end;

procedure TfrmSpotEdit.ScrollBox1MouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 if Button=mbLeft then
 begin
  if FDrawing and (not FOutOfControl) then
   AddPoint(DtoI(FPaintBox.ScreenToClient(ScrollBox1.ClientToScreen(point(X,Y)))));

 end;
end;

procedure TfrmSpotEdit.edSpotIDKeyPress(Sender: TObject; var Key: Char);
begin
 if not (Key  in [#8,'0'..'9']) then
 Key:=#0;
end;

procedure TfrmSpotEdit.rgZoomClick(Sender: TObject);
var
 Denominator,Nominator:integer ;
begin
 case rgZoom.ItemIndex of
 0:{50%}
   begin
    Nominator:=1;
    Denominator:=2;
   end;
 1:{100%}
   begin
    Nominator:=1;
    Denominator:=1;
   end;
 2:{200%}
  begin
    Nominator:=2;
    Denominator:=1;
   end;
 3:{300%}
   begin
    Nominator:=3;
    Denominator:=1;
   end;
 else
    Nominator:=1;
    Denominator:=1;
 end;
 SetScale(Nominator,Denominator);
end;

end.
